#ifndef SUPPORT_EXIT_H
#define SUPPORT_EXIT_H

#define EXIT_MLIF_BASE (0x10)
#define EXIT_MLIF_INVALID_SIZE (EXIT_MLIF_BASE + 1)
#define EXIT_MLIF_MISSMATCH (EXIT_MLIF_BASE + 2)

#endif  // SUPPORT_EXIT_H
