---
# The MLONMCU_HOME is filled in automatically when creating the environment
home: "{{ home_dir }}"
logging:
  level: DEBUG
  to_file: false
  rotate: false
cleanup:
  auto: true
  keep: 50
# Default locations for certain directoriescan be changed here
# Non-absolute paths will always be threated relative to the MLONMCU_HOME
paths:
  # Where the dependencies are downloaded and installed
  deps: deps
  # If logging to file is used keep logs in this directory
  logs: logs
  # Location where reports and artifacts are written to
  results: results
  # Directory where custom extensions can be integrated
  plugins: plugins
  # Directory for intermediate build products, should be located on a large enough drive
  temp: temp
  # A collection of models which will be used to look for models
  # The paths will be checked in the order defined here stopping at the first match
  # Non-existant paths will be skipped without throwing an error
  models:
    - "{{ home_dir }}/models"
    - "{{ config_dir }}/models"
# Here default clone_urls
repos:
  tensorflow:
    url: "https://github.com/tensorflow/tflite-micro.git"
    ref: f050eec7e32a0895f7658db21a4bdbd0975087a5
  tvm:
    url: "https://github.com/apache/tvm.git"
    ref: de6d8067754d746d88262c530b5241b5577b9aae
    options:
      recursive: true
  utvm_staticrt_codegen:
    url: "https://github.com/tum-ei-eda/utvm_staticrt_codegen.git"
    ref: 1b296f1efcc28671ee21ccbe86a81b968db8b1e5
  tvm_extensions:
    url: "https://github.com/tum-ei-eda/tvm_extensions.git"
    ref: 3ccd1ad8e4eed0cd86bfbd866d066833fd174dcb
  muriscvnn:
    url: "https://github.com/tum-ei-eda/muriscv-nn.git"
    ref: c023b80a51c1b48ec62b9b092d047e9ac0bab3e8
  etiss:
    url: "https://github.com/tum-ei-eda/etiss.git"
    ref: 40c0b920cf4072adf189211ca44291534355dcfd
  spike:
    url: "https://github.com/riscv-software-src/riscv-isa-sim.git"
    ref: 0bc176b3fca43560b9e8586cdbc41cfde073e17a
  spikepk:
    url: "https://github.com/riscv-software-src/riscv-pk.git"
    ref: 7e9b671c0415dfd7b562ac934feb9380075d4aa2
  cmsis:
    url: "https://github.com/PhilippvK/CMSIS_5.git"
    ref: ad1c3cad8f1240ef14a2c55381a78d792d76ec4d
  mlif:
    url: "https://github.com/tum-ei-eda/mlonmcu-sw.git"  # This should be later changed to main branch in mlonmcu-sw
    ref: 4b9a32659f7c5340e8de26a0b8c4135ca67d64ac
  espidf:
    url: "https://github.com/espressif/esp-idf.git"
    ref: release/v4.4
    options:
      recursive: true
  microtvm_etiss:
    url: "https://github.com/PhilippvK/microtvm-etiss-template.git"
    ref: 95ed42e4cc1dfbcccb39a0a5b8f95ca22bbef9cc
  zephyr:
    url: "https://github.com/zephyrproject-rtos/zephyr.git"
    ref: v3.2.0
  pulp_gcc:
    url: "https://github.com/pulp-platform/riscv-gnu-toolchain.git"
    ref: 11ba51efe1c4692d60357083b9c20c28bfb66a23
    options:
      recursive: true
  pulp_freertos:
    url: "https://github.com/pulp-platform/pulp-freertos.git"
    ref: aabd624fd9feef222f8dc2e0fa3b7dc487c42bf3
    options:
      recursive: true
# Here all supported frameworks with their specific features are defined
# Optionally disable unwanted or incomatible backends or features here
# The configured defaults are used if no backend was specified in the command line options
frameworks:
  default: tvm
  tflm:
    enabled: true
    backends:
      default: tflmi
      tflmi:
        enabled: true
        features:
          debug_arena: true
    features:
      muriscvnn: true
      cmsisnn: false
  tvm:
    enabled: true
    backends:
      default: tvmaot
      tvmrt:
        enabled: true
        features:
          debug_arena: true
          disable_legalize: false
          autotuned: true
          moiopt: true
      tvmaot:
        enabled: true
        features:
          debug_arena: true
          unpacked_api: true
          usmp: true
          disable_legalize: false
          autotuned: true
          moiopt: true
      tvmaotplus:
        enabled: true
        features:
          debug_arena: true
          disable_legalize: false
          autotuned: true
          moiopt: true
      tvmllvm:
        enabled: true
        features:
          # unpacked_api: true
          disable_legalize: false
          autotuned: true
          moiopt: true
      tvmcg:
        enabled: true
        features:
          debug_arena: true
          disable_legalize: false
          autotuned: true
    features:
      cmsisnnbyoc: false
      muriscvnnbyoc: true
# Some frontends are experimental and therefore disabled here
# Features like packing are only available in certain environments
frontends:
  tflite:
    enabled: true
    features:
      validate: true
      visualize: true
  relay:
    enabled: true
    features:
      relayviz: true
  packed:
    enabled: false
    features:
      packing: true
      packed: true
  onnx:
    enabled: true
  pb:
    enabled: true
  paddle:
    enabled: true
  # TODO: saved_model (TF->TFLITE), ipynb (IPYNB->?)
# Some targets/platforms support multiple toolchains
toolchains:
  gcc: true
  llvm: true
# Platform extend the number of supported targets
platforms:
  mlif:
    enabled: true
    features:
      debug: true
      validate: true
      benchmark: true
  espidf:
    enabled: false
    features:
      debug: true
  zephyr:
    enabled: false
    features:
      debug: true
  tvm:
    enabled: true
    features:
      benchmark: true
      tvm_rpc: true
      autotune: true
      tvm_profile: true
  microtvm:
    enabled: true
    features: []
      # validate: true
# List of supported targets in the environment
targets:
  default: etiss_pulpino
  etiss_pulpino:
    enabled: true
    features:
      gdbserver: true
      etissdbg: true
      trace: true
      log_instrs: true
      # vext: true
      # pext: true
  host_x86:
    enabled: true
    features:
      gdbserver: true
  spike:
    enabled: true
    features:
      vext: false
      pext: false
      cachesim: true
      log_instrs: true
  ovpsim:
    enabled: false
    features:
      vext: false
      pext: false
      log_instrs: true
  corstone300:
    enabled: false
    features:
      ethosu: false
      arm_mvei: true
      arm_dsp: true
  riscv_qemu:
    enabled: true
    features:
      vext: false
  microtvm_host:
    enabled: true
  microtvm_spike:
    enabled: true
  microtvm_zephyr:
    enabled: false
  microtvm_arduino:
    enabled: false
  microtvm_etiss:
    enabled: true
  microtvm_espidf:
    enabled: false
  tvm_cpu:
    enabled: true
  gvsoc_pulp:
    enabled: true
  gvsoc_pulpissimo:
    enabled: true
postprocesses:
  use:
  # - detailed_cycles
  # - average_cycles
  # - filter_cols
  # - features2cols
  # - config2cols
  # - bytes2kb
  # - visualize
# This is where further options such as specific versions of dependencies can be set in the furture
vars:
  allow_extensions: false
  llvm.version: "14.0.0"
  runs_per_stage: true
  pulp_gcc.dl_url: "https://github.com/pulp-platform/pulp-riscv-gnu-toolchain/releases/download/v1.0.16/v1.0.16-pulp-riscv-gcc-ubuntu-18.tar.bz2"
  # pulp_gcc.from_source: true
  # pulp_freertos.experimental_install: true
flags:
  tflmc.exe:
  - x86
