/*
 * Copyright (c) 2022 TUM Department of Electrical and Computer Engineering.
 *
 * This file is part of MLonMCU.
 * See https://github.com/tum-ei-eda/mlonmcu.git for further info.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#ifndef SUPPORT_EXIT_H
#define SUPPORT_EXIT_H

#define EXIT_MLIF_BASE        (0x10)
#define EXIT_MLIF_INVALID_SIZE  (EXIT_MLIF_BASE + 1)
#define EXIT_MLIF_MISSMATCH   (EXIT_MLIF_BASE + 2)

#endif  // SUPPORT_EXIT_H
